# Nexus - the Nuclear Elastic X-ray scattering Universal Software package
# 
# Copyright 2020 - 2023 Deutsches Elektronen-Synchrotron DESY
# A research centre of the Helmholtz Association.
# All rights reserved.
#
# Author: Lars Bocklage - lars.bocklag@desy.de
# 
# This file is part of Nexus.
# 
# Nexus is free software; you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# 
# Nexus is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with Nexus.
# If not, see <https:www.gnu.org/licenses/>.

 
"""
Module for data handling.
"""

from nexus import clib
from nexus.lib import distribution
from nexus.lib import material
from nexus.lib import moessbauer

import numpy as np
from scipy.optimize import least_squares
from scipy.interpolate import interp1d
from scipy.signal import find_peaks

from pybaselines.polynomial import poly
from pybaselines import Baseline as pyBase


def Save(file_name, data_list):
    r"""
    Save data arrays as columns to file.
    
    Args:
       file_name (string): File to load (with relative path).
       data_list (list): List containing the data to save column wise.
    """
    np.savetxt(file_name, np.column_stack(data_list))


def Load(file_name, x_index, intensity_index, x_start = None, x_stop = None, intensity_threshold = None):
    r"""
    Load data arrays from columns in file.
    
    Args:
       file_name (string): File to load (with relative path).
       x_index (int): Column index in file for x axis data.
       intensity_index (int): Column index in file for intensity data.
       x_start (float): x data smaller than x_start are not considered.
       x_stop (float): x data larger than x_stop are not considered.
       intensity_threshold (float): intensity data <= intensity_threshold are not considered.
    """
    data = np.loadtxt(file_name)
    
    x = data[:,x_index]
    intensity = data[:,intensity_index]

    mask_x_start = np.ones(len(x))
    if x_start != None:
        mask_x_start = np.where(x < x_start, False, True)
    
    mask_x_stop = np.ones(len(x))
    if x_stop != None:
        mask_x_stop = np.where(x > x_stop, False, True)
    
    mask_int = np.ones(len(intensity))
    if intensity_threshold != None:
        mask_int = np.where(intensity <= intensity_threshold, False, True)

    mask = mask_x_start * mask_x_stop * mask_int
    
    x = x[np.nonzero(mask)]

    intensity = intensity[np.nonzero(mask)]
    
    return x, intensity


def HistogramStdDev(data):
    r"""
    Determines the standard deviation of histogram data, which is given by the square root of the measured histogram value.
    
    For Poisson statistics the expectation value (here the one single measured value in a data set) and the variance are the same.
    Thus, the variance is :math:`Var = y = \sigma_y^2`, with the measured value :math:`y`.
    The standard deviation is :math:`\sigma_y = \sqrt{y}`.
    """
    return np.sqrt(data)


def Binning(data, bin=2, method="sum" , endpoint=False):
    r"""
    Performs a binning/bucketing procedure on the data array with the given bin size.
    Either the sum or the mean of the binned data points can be calculated.
    If data array cannot be divided into equal sizes the endpoint can be discarded or binned with smaller size.

    Args:
      data (list): Data to be folded.
      bin (int): Bin size, the number of points to be processed.
      method (string): Either ``sum`` or ``mean``.
      endpoint (bool): Determines the endpoint treatment in case the array length is not integer dividable by bin.
        If ``False`` the last points are discarded.
        If ``True`` the bin of the last points equals the number of points left.
        It will result in non-equidistant points on the x-axis.

    Returns:
      ndarray: Binned data.
    """
    if bin == 1:
        return data
    
    if method not in ["sum", "mean"]:
        raise Exception("method not recognized")
    
    if endpoint:
        points = int(np.ceil(len(data) / bin))
    else:
        points = int(np.floor(len(data) / bin))
        
    bin_data = np.zeros(points)
    
    for i in range(len(bin_data)):
        step = i * bin
        
        elem = data[step:step+bin]
        
        if method == "sum":
            bin_data[i] = np.sum(elem)
        elif method == "mean":
            bin_data[i] = np.mean(elem)
    
    return bin_data


def Fold(data, folding_point, flip='left'):
    r"""
    Folds a one dimensional data array at the given folding point.
    The given array is cut in half.
    Data of the second half are rolled and re-introduced on the other side to ensure equal intensities.
    Give an even number of data points.

    The central shift depends on the folding direction (left or right).
    The folding direction depends on the experimental setup.
    Measure the second order Doppler shift to determine the correct folding.

    .. versionchanged:: 1.0.2

    Args:
      data (list): Data to be folded. Number of data points has to be even.
      folding_point (int): Folding point of the array.
      flip (string): determines whether the left or right part of the spectrum is flipped.
        
        .. versionadded:: 1.0.2

    Returns:
      ndarray: Folded data.
    """
    num_points = len(data) / 2

    split_data = np.hsplit(np.array(data), 2)

    if flip == 'right':
        left_data = split_data[0]

        right_data = np.flip(split_data[1])
    elif flip == 'left':
        left_data = np.flip(split_data[0])

        right_data = split_data[1]
    
    shift_value = int(folding_point - num_points)

    right_data = np.roll(right_data, shift_value)
    
    return left_data + right_data


def AutoFold(data, flip='left'):
    r"""
    Folds a one dimensional data array.
    Uses autocorrelation to determine the correct folding point.
    The given data array is cut in half.
    Data of the second half are rolled and re-introduced on the other side to ensure equal intensities.
    Give an even number of data points.
    Returns the folded data and the fold index.

    The central shift depends on the folding direction (left or right).
    The folding direction depends on the experimental setup.
    Measure the second order Doppler shift to determine the correct folding.

    .. versionchanged:: 1.0.2

    Args:
      data (list): Data to be folded. Number of data points has to be even.
      flip (string): determines whether the left or right part of the spectrum is flipped.
        
        .. versionadded:: 1.0.2

    Returns:
      ndarray, int: Folded data, fold_index
    """
    split_data = np.hsplit(np.array(data), 2)

    if flip == 'right':
        left_data = split_data[0]

        right_data = np.flip(split_data[1])
    elif flip == 'left':
        left_data = np.flip(split_data[0])

        right_data = split_data[1]

    correlation = np.correlate(left_data, right_data, mode = 'same')

    correlation_maximum = np.argmax(correlation)
    
    shift_value = int(correlation_maximum - (len(left_data)) / 2)

    right_data = np.roll(right_data, shift_value)
    
    return left_data + right_data, int(len(data)/2 + shift_value)


def Fold_2D_hor(data, flip='left'):
    split_data = np.hsplit(np.array(data), 2)

    if flip == 'right':
        left_data = split_data[0]

        right_data = np.flip(split_data[1], axis=1)
    elif flip == 'left':
        left_data = np.flip(split_data[0], axis=1)

        right_data = split_data[1]
    
    correlation = np.correlate(np.sum(left_data, axis=0), np.sum(right_data, axis=0), mode = 'same')

    correlation_maximum = np.argmax(correlation)
    
    shift_value = int(correlation_maximum - (left_data.shape[1]) / 2)

    right_data = np.roll(right_data, shift_value, axis=1)

    return left_data + right_data, int(data.shape[1] / 2 + shift_value)


def Fold_2D_vert(data, flip='left'):
    split_data = np.vsplit(np.array(data), 2)

    if flip == 'right':
        left_data = split_data[0]

        right_data = np.flip(split_data[1], axis=0)
    elif flip == 'left':
        left_data = np.flip(split_data[0], axis=0)

        right_data = split_data[1]
    
    correlation = np.correlate(np.sum(left_data, axis=1), np.sum(right_data, axis=1), mode = 'same')

    correlation_maximum = np.argmax(correlation)
    
    shift_value = int(correlation_maximum - (left_data.shape[0]) / 2)

    right_data = np.roll(right_data, shift_value, axis=0)

    return left_data + right_data, int(data.shape[0] / 2 + shift_value)


def AutoFold_2D(data, axis='hor', flip = 'left'):
    r"""
    Folds a two dimensional data array along the given axis.
    Uses autocorrelation on the data, which are integrated over the second axis, to determine the correct folding point.
    The given data array is cut in half.
    Data of the second half are rolled and re-introduced on the other side to ensure equal intensities.
    Give an even number of data points along the fold direction.
    Returns the folded data and the fold index.

    The central shift depends on the folding direction (left or right).
    The folding direction depends on the experimental setup.
    Measure the second order Doppler shift to determine the correct folding.
    
    .. versionchanged:: 1.0.2

    Args:
      data (list): Data to be folded. Number of data points has to be even along fold direction.
      axis (int): Either *hor* (horizontal) or *vert* (vertical) folding.
      flip (string): determines whether the left or right (upper/lower) part of the spectrum is flipped.
        
        .. versionadded:: 1.0.2

    Returns:
      ndarray, int: Folded data, fold index
    """
    result = data

    if axis == 'hor':
      result, shift_value = Fold_2D_hor(data, flip)
    elif axis == 'vert':
      result, shift_value = Fold_2D_vert(data, flip)
      
    return result, shift_value


def ChannelsToVelocity(num_channels, velocity, offset=0, mode="constant"):
    r"""
    Converts a given number of channels to a velocity grid with possible offset for either sinusoidal mode or constant acceleration mode of a MOessbauer drive unit.

    Args:
      num_channels (int): Number of channels.
      velocity (float): maximum velocity in mm/s.
        The velocity is set symmetrically around zero.
      offset (float): Offset to the symmetric velocity in mm/s. default is 0.
      mode (string): Either constant acceleration mode (``constant``) or sinusoidal mode (``sinus``) of the Moessbauer drive unit.
        Default is ``constant``.

    Returns:
      ndarray: Velocity of the channels in mm/s.
    """
    if mode == "constant":
        velocities = np.linspace(-velocity, velocity, num_channels)
    elif mode == "sinus":
        channels = np.arange(num_channels)
        velocities = -velocity * np.cos(channels / len(channels) * np.pi)
    else:
        raise Exception("mode not recognized.")

    velocities = velocities + offset

    return velocities


def CalibrateChannels(intensity, velocity, intensity_reference, velocity_reference, mode="constant", shift = 0.0, center_shift = 0.0):
    r"""
    Calibrates the velocity of the intensity data set to a given reference data set with known velocity.
    The intensity array is fit against the reference spectrum.

    .. versionchanged:: 1.0.3
       The velocity offset is now applied to the theory not to the experiment anymore.

    .. versionchanged:: 1.0.4
       Reverted change of version 1.0.3. Offset is applied to velocity again to properly account for drive offset.

    Args:
      intensity (list): Data for which the velocity should be calibrated.
      velocity (float or list): If a float is given, the velocity is the assumed maximum velocity of the data set.
        If a list is provided, it is the assumed experimental velocity grid.
      intensity_reference (list): Array of intensity data used as reference. 
      velocity_reference (list): Array of velocities used as reference.
      mode (string): Either constant acceleration mode (``constant``) or sinusoidal mode (``sinus``) of the Moessbauer drive unit. 
        Default is ``constant``.
      shift (float): Initial guess for shift of the spectrum in mm/s. Optional, default is zero.
        
        .. versionadded:: 1.0.3

      center_shift (float): Offset to the velocity calibration in mm/s. Optional, default is zero.
        For example, when a center shift is known between source and velocity reference sample.

        .. versionadded:: 1.0.4

    Returns:
      ndarray, float, float, ndarray, ndarray:
         calibrated experimental velocity,
         velocity offset,
         velocity scaling factor,
         velocity used for the theoretical spectrum,
         theoretical spectrum
    """
 
    if isinstance(velocity, (int, float)):
        if mode == "constant":
            velocities = np.linspace(-velocity, velocity, len(intensity))
        elif mode == "sinus":
            channel = np.arange(len(intensity))
            velocities = -velocity * np.cos(channel / len(channel) * np.pi)
        else:
            raise Exception("mode not recognized.")
    elif isinstance(velocity, (list, tuple, np.ndarray)):
        velocities = velocity
    else:
        raise Exception("velocity has wrong format.")
    
    # this is a pointer to an interpolation function for the theory data
    theory_interpolation_function = interp1d(velocity_reference, intensity_reference, fill_value = 'extrapolate')
    
    def fit_func(x, intensity, velocities, theory_interpolation_func):
        velocity_exp = x[1] * velocities - x[0]
        
        return intensity - (x[3] * theory_interpolation_func(velocity_exp) + x[2])

    result = least_squares(fit_func,
                           x0 = [shift, 1, np.mean(intensity), max(intensity)-min(intensity)], # velocity_offset, velocity_scaling, int_offset, int_scaling
                           args = (intensity, velocities, theory_interpolation_function),
                           method = 'trf')
    
    velocity_exp =  result.x[1] * velocities - result.x[0] - center_shift

    intensity_reference_scaled = result.x[3] * intensity_reference + result.x[2]

    return velocity_exp, result.x[0], result.x[1], velocity_reference, intensity_reference_scaled



def CalibrateChannelsExperiment(intensity, velocity, experiment, emission=False, mode="constant", shift = 0.0, center_shift = 0.0):
    r"""
    Calibrates the velocity of a reference experiment.
    The theoretical model of the reference :class:`Experiment` must be given.
    The intensity array is fit against the theoretical spectrum.

    .. versionchanged:: 1.0.3
       The velocity offset is now applied to the theory not to the experiment anymore.

    .. versionchanged:: 1.0.4
       Reverted change of version 1.0.3. Offset is applied to velocity again to properly account for drive offset.

    Args:
      intensity (list): Data for which the velocity should be calibrated.
      velocity (float or list): If a float is given, the velocity is the assumed maximum velocity of the data set.
        If a list is provided, it is the assumed experimental velocity grid.
      experiment (nx.Experiment): An :class:`Experiment` against which the experimental velocity is fit.
      emission (bool): If ``False`` (default), an absorption spectrum is calculated. If ``True`` an emission spectrum is calculated.
      mode (string): Either constant acceleration mode (``constant``) or sinusoidal mode (``sinus``) of the Moessbauer drive unit. 
        Default is ``constant``.
      shift (float): initial guess for shift of the spectrum in mm/s.

        .. versionadded:: 1.0.3

      center_shift (float): Offset to the velocity calibration in mm/s. Optional, default is zero.
        For example, when a center shift is known between source and velocity reference sample.

        .. versionadded:: 1.0.4

    Returns:
      ndarray, float, float, ndarray, ndarray:
         calibrated experimental velocity,
         velocity offset,
         velocity scaling factor,
         velocity used for the theoretical spectrum,
         theoretical spectrum
    """
    if isinstance(velocity, (int, float)):
        v_max = velocity
    elif isinstance(velocity, (list, tuple, np.ndarray)):
        v_max = max(velocity, key=abs)
    else:
        raise Exception("velocity has wrong format.")

    # theory calculations are performed on linear grid
    velocities_theory = np.linspace(-v_max, v_max, len(intensity))

    if (emission == True):
        spectrum = clib.EmissionSpectrum(experiment = experiment,
                                         velocity = velocities_theory)
    else:
        spectrum = clib.MoessbauerSpectrum(experiment = experiment,
                                           velocity = velocities_theory)
        
    spectrum_theory = spectrum()
    
    return CalibrateChannels(intensity, velocity, spectrum_theory, velocities_theory, mode, shift = shift, center_shift = center_shift)


def CalibrateChannelsAlphaFe(intensity, velocity, thickness, Bhf=33, B_fwhm=0.3, emission=False, mode="constant", shift = 0.0, center_shift = 0.0):
    r"""
    Calibrates the velocity of an isotropically distributed :math:`\\alpha`-Fe spectrum.
    No beam polarization.
    The intensity array is fit against a theoretical sextet of :math:`\\alpha`-Fe.

    .. versionchanged:: 1.0.3
       The velocity offset is now applied to the theory not to the experiment anymore.

    .. versionchanged:: 1.0.4
       Reverted change of version 1.0.3. Offset is applied to velocity again to properly account for drive offset.

    Args:
      intensity (list): Data from an :math:`\\alpha`-Fe sample.
      velocity (float or list): If a float is given, the velocity is the assumed maximum velocity of the data set.
        If a list is provided, it is the assumed experimental velocity grid.
      thickness (float): Thickness of the :math:`\\alpha`-Fe sample (nm).
        Please note, that for sample thicknesses exceeding the typical escape length of the emitted probe (X-ray fluorescence or electrons) the escape length should be given.
      Bhf (float): The magnitude of the magnetic hyperfine field.
        Default is 33 T.
      B_fwhm (float): The FWHM of the magnetic hyperfine field distribution.
        Default is 0.3 T.
      emission (bool): If ``False`` (default), an absorption spectrum is calculated. If ``True`` an emission spectrum is calculated.
      mode (string): Either constant acceleration mode (``constant``) or sinusoidal mode (``sinus``) of the Moessbauer drive unit. 
        Default is ``constant``.
      shift (float): initial guess for shift of the spectrum in mm/s.

        .. versionadded:: 1.0.3

      center_shift (float): Offset to the velocity calibration in mm/s. Optional, default is zero.
        For example, when a center shift is known between source and velocity reference sample.

        .. versionadded:: 1.0.4

    Returns:
      ndarray, float, float, ndarray, ndarray:
          calibrated experimental velocity,
          velocity offset,
          velocity scaling factor,
          velocity used for the theoretical spectrum,
          theoretical spectrum
    """
    dist = distribution.Gaussian(points = 31, 
                                 fwhm = B_fwhm)
    
    site = clib.Hyperfine(magnetic_field = Bhf,
                          magnetic_field_dist = dist,
                          isotropic = True)
    
    mat = clib.Material.Template(material.Fe)
    
    mat.hyperfine_sites = [site]
    
    layer = clib.Layer(thickness = thickness,
                     material = mat)
    
    sample = clib.Sample(layers = [layer])
    
    beam = clib.Beam()
    beam.Unpolarized()
    
    exp = clib.Experiment(beam = beam,
                          objects = [sample],
                          isotope = moessbauer.Fe57)

    return CalibrateChannelsExperiment(intensity, velocity, experiment = exp, emission = emission, mode = mode, shift = shift, center_shift = center_shift)


def FindMinima(x_data, y_data, n=None):
    r"""
    Returns the x and y values of minima or of a single minimum found on the y data set.

    .. versionadded:: 1.0.2

    Args:
      x_data (list): x data.
      y_data (list): y data on which the minima are searched for.
      n (int): returns only the n-th minimum, optional parameter.

    Returns:
      ndarray or float, ndarray or float, ndarray or int: minima value(s) on the x axis, minima value(s) on the y axis, indices of the minima
    """
    indices, _ = find_peaks(-y_data)

    if n is not None:
        indices = indices[n-1]

    return x_data[indices], y_data[indices], indices


def FindMaxima(x_data, y_data, n=None):
    r"""
    Returns the x and y values of maxima or of a single maximum found on the y data set.

    .. versionadded:: 1.0.2

    Args:
      x_data (list): x data.
      y_data (list): y data on which the minima are searched for.
      n (int): returns only the n-th minimum, optional parameter.

    Returns:
      ndarray or float, ndarray or float, ndarray or int: minima value(s) on the x axis, minima value(s) on the y axis, indices of the minima
    """

    indices, _ = find_peaks(y_data)

    if n is not None:
        indices = indices[n-1]

    return x_data[indices], y_data[indices], indices


def BaselinePoly(data, left_point, right_point, poly_order = 2):
    r"""
    Apply a polynomial baseline correction to the data.
    In order to correct the baseline, only parts of the spectrum should be used where no peaks are visible.

    Be aware, that baseline correction changes the intensity of the lines.

    For more options and other algorithms use the `pybaselines` package.

    .. versionadded:: 1.0.4

    Args:
      data (list): Data to be corrected.
      left_point (int): Index of the data array. Data from zero to this index are used.
      left_point (int): Index of the data array. Data from this point to the end of the array are used.
      poly_order (int): Polynominal order for fitting the baseline. Default is 2.

    Returns:
      ndarray, ndarray: Returns the data corrected by the baseline and the baseline.
    """
    data = np.array(data)

    x = np.arange(len(data))

    non_peaks = ((x < left_point) | (x > right_point))
    x_masked = x[non_peaks]
    d_masked = data[non_peaks]

    _, params = poly(d_masked, x_masked, poly_order=poly_order, return_coef=True)
    
    baseline = np.polynomial.Polynomial(params['coef'])(x)

    return data - baseline + np.mean(baseline), baseline


def Baseline(data, lam, algorithm = 'aspls', max_iter=100, tol=0.001):
    r"""
    Apply various algorithms for baseline correction of the data.
    The correct ``lam`` value has to be specified by the user.
    A good procedure is to start with a low value (100) and then increase in an exponential manner (1e3, 1e4, ...).
    At some point the curvature of the baseline changes.
    Then, reduce the values again and adjust.

    Be aware, that baseline correction changes the intensity of the lines.

    For more options and other algorithms use the `pybaselines` package.

    .. versionadded:: 1.0.4

    Args:
      data (list): Data to be corrected.
      lam (float): The smoothing parameter. Larger values will create smoother baselines.
      algorithm (string): Defines the used algorithm.

        * ``aspls`` Adaptive Smoothness Penalized Least Squares.
        * ``mpls`` Morphological Penalized Least Squares.
        * ``asls`` Asymmetric Least Squares.
        * ``drlps`` Doubly Reweighted Penalized Least Squares.
        * ``arpls`` Asymmetrically Reweighted Penalized Least Squares.
        * ``iarpls`` Improved Asymmetrically Reweighted Penalized Least Squares.
        * ``amormol`` Iteratively averaging morphological and mollified baseline fit.
      
        See `pybaselines` package documentaiton for more information.

      max_iter (int): Maximum number of iterations. Default is 100.
      tol (float): Tolerance value, defines the exit criteria. Default is 0.001.

    Returns:
      ndarray, ndarray: Returns the data corrected by the baseline and the baseline.
    """
    data = np.array(data)

    x = np.arange(len(data))

    baseline_fitter = pyBase(x_data=x)

    if algorithm == 'aspls':
        baseline, _ = baseline_fitter.aspls(data, lam=lam, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'mpls':
        baseline, _ = baseline_fitter.mpls(data, half_window=None, lam=lam, p=0.0, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'asls':
        baseline, _ = baseline_fitter.asls(data, lam=lam, p=0.01, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'drpls':
        baseline, _ = baseline_fitter.drpls(data, lam=lam, eta=0.5, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'arpls':
        baseline, _ = baseline_fitter.arpls(data, lam=lam, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'iarpls':
        baseline, _ = baseline_fitter.iarpls(data, lam=lam, diff_order=2, max_iter=max_iter, tol=tol, weights=None)
    elif algorithm == 'amormol':
        baseline, _ = baseline_fitter.amormol(data, half_window=None, max_iter=max_iter, tol=tol, pad_kwargs=None)
    else:
        raise Exception('Nexus warning - algorithm in baseline_Wittaker not found.')
        
    return data - baseline + np.mean(baseline), baseline