# Nexus - the Nuclear Elastic X-ray scattering Universal Software package
# 
# Copyright 2020 - 2023 Deutsches Elektronen-Synchrotron DESY
# A research centre of the Helmholtz Association.
# All rights reserved.
#
# Author: Lars Bocklage - lars.bocklag@desy.de
# 
# This file is part of Nexus.
# 
# Nexus is free software; you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# 
# Nexus is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with Nexus.
# If not, see <https:www.gnu.org/licenses/>.

 
"""
Predefined residuals derived from the parent class :class:`Residual`.
"""

import numpy as np

from nexus import clib


class Sqrt(clib.Residual):
    r"""
    Difference of the square root values
    
    .. math::

       residual = \sqrt{data} - \sqrt{theory}

    This an approximation of the Poisson likelihood function [Thibault]_. and is the standard residual for most :class:`FitMeasurement` types (if the residual parameter to ``None``).
    """
    def __init__(self, id = "Sqrt", plot_string = r"$\sqrt{x} - \sqrt{x_t}$"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        #sqrt_data = np.where(np.array(input_data) <= 0.0, 0.0, np.sqrt(input_data))
        #sqrt_theory = np.where(np.array(input_theory) <= 0.0, 0.0, np.sqrt(input_theory))
        sqrt_data = [np.sqrt(x) if x>=0.0 else 0.0 for x in input_data]
        sqrt_theory = [np.sqrt(x) if x>=0.0 else 0.0 for x in input_theory]
        return np.subtract(sqrt_data, sqrt_theory)


class StdDev(clib.Residual):
    r"""
    Difference residual weighted by the standard deviation
    
    .. math::

       residual = \frac{data - theory}{\sqrt{data}}

    For histograms the variance is the measured data value :math:`V = data` and the standard deviation is given by :math:`\sigma = \sqrt(V) = \sqrt(data)`.
    The best linear unbiased estimator for Gaussian noise is obtained by using the variance as a weight in the squared residual.
    This will result in the cost function
    
    .. math::

       cost = \frac{1}{2} \sum_i \frac{r_i^2}{V_i}  = \frac{1}{2} \sum_i \frac{r_i^2}{\sigma_i^2} = \frac{1}{2} \sum_i \frac{(data_i- theory_i)^2}{|data_i|}

    This is a very typical cost function for data fitting.
    However, for Poisson statistics and low count rates a Poisson likelihood function is the better choice.

    .. seealso:: `<https://en.wikipedia.org/wiki/Weighted_least_squares>`_
    """
    def __init__(self, id = "StdDev", plot_string = r"$\dfrac{x - x_t}{\sqrt{x}}$"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        sqrt_data = np.where(np.array(input_data) <= 0.0, 1.0, np.sqrt(input_data))
        return np.subtract(input_data, input_theory) / sqrt_data


class Difference(clib.Residual):
    r"""
    Difference residual
    
    .. math::
    
       residual = data - theory

    """
    def __init__(self, id = "Difference", plot_string = r"$x - x_t$"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        return np.subtract(input_data, input_theory)


class Norm(clib.Residual):
    r"""
    Normalized difference residual
    
    .. math::

       residual = \frac{data - theory}{\sum data}

    """
    def __init__(self, id = "Norm", plot_string = r"\dfrac{x - x_t}{\sum x}"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        sum = np.sum(input_data)
        if sum == 0:
            sum = 1
        return np.subtract(input_data, input_theory) / sum


class Log10(clib.Residual):
    r"""
    Difference of the logarithm of base 10 values

    .. math::

       residual = \log_{10}(data) - \log_{10}(theory)

    """
    def __init__(self, id = "Log10", plot_string = r"$\log_{10}(x) - \log_{10}(x_t)$"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        #log_data = np.where(np.array(input_data) > 1.0001, np.log10(input_data), 0.00001)
        #log_theory = np.where(np.array(input_theory) > 1.0001, np.log10(input_theory), 0.00001)
        log_data = [np.log10(x) if x>1.0 else 1e-10 for x in input_data]
        log_theory = [np.log10(x) if x>1.0 else 1e-10 for x in input_theory]
        return np.subtract(log_data, log_theory)


class Ln(clib.Residual):
    r"""
    Difference of the natural logarithm values

    .. math::

       residual = \ln(data) - \ln(theory)

    """
    def __init__(self, id = "Ln", plot_string = r"$\ln(x) - \ln(x_t)$"):
        super().__init__(id, plot_string)

    def ResidualFunction(self, input_data, input_theory):
        #ln_data = np.where(np.array(input_data) > 1.001, np.log(input_data), 0.000001)
        #ln_theory = np.where(np.array(input_theory) > 1.001, np.log(input_theory), 0.000001)
        # avoid division by zero error
        ln_data = [np.log(x) if x>1 else 1e-10 for x in input_data]
        ln_theory = [np.log(x) if x>1 else 1e-10 for x in input_theory]
        return np.subtract(ln_data, ln_theory)
